"""
Tools for creating simulated images with noise,
for conceptual signal-to-noise demonstrations.
"""
from .imports import *
from .psfs import *
from chromatic.rainbows.visualizations.utilities import (
    get_animation_writer_and_displayer,
)
from IPython.display import display
from photutils.aperture import (
    CircularAperture,
    aperture_photometry,
    ApertureStats,
    CircularAnnulus,
)


def to_photons(x):
    '''
    Wrapper to cleanly convert electrons to photons, if necessary.
    '''
    if "electron" in x.unit.to_string():
        return x * u.photon / u.electron
    elif "photon" in x.unit.to_string():
        return x


class Image:
    '''
    
    '''
    keys = ["shape", "gain", "pixel_scale", "psf"]

    def __init__(
        self,
        components=[],
        shape=(30, 30),
        gain=2 * u.photon / u.adu,
        pixel_scale=0.5 * u.arcsec / u.pixel,
        psf=seeing(fwhm=2.0 * u.arcsec),
    ):
        '''
        Initialize a simulated astronomical image. 
        '''
        self.shape = shape
        self.gain = gain
        self.pixel_scale = pixel_scale
        self.psf = psf
        self.components = components
        for c in self.components:
            c.image = self

    def describe_as_string(self):
        '''
        Describe the basics of this image. 
        '''
        s = ""  # f"{self.__class__.__name__}:\n"
        for k in self.keys:
            s += f"{k}={getattr(self, k)}\n"
        return s

    def describe_everything_as_string(self):
        '''
        Describe all the components of this image. 
        '''
        return "".join([c.describe_as_string() for c in [self] + self.components])

    def do_photometry(
        self,
        positions=[(15, 15)],
        radius=3,
        sky_subtract=True,
        radius_sky_inner=6,
        radius_sky_outer=10,
        visualize=True,
        **kwargs,
    ):
        '''
        Do photometry on a star in this image. 
        '''
        self.aperture = CircularAperture(positions=positions, r=radius)

        self.photometry_results = aperture_photometry(
            data=self.images["data"],
            apertures=self.aperture,
            error=self.images["uncertainty"],
        )

        flux = self.photometry_results["aperture_sum"].data[0]
        uncertainty = self.photometry_results["aperture_sum_err"].data[0]
        self.aperture_photometry = aperture_photometry(
            data=self.images["data"],
            apertures=self.aperture,
            error=self.images["uncertainty"],
        )
        print(f"{self.aperture.area:.1f} pixel = ⭕️ aperture area")
        print(f"{flux:>10.1f} photons = ⭐️ measured fluence in aperture")
        flux_model = aperture_photometry(
            data=self.images["model"],
            apertures=self.aperture,
        )["aperture_sum"].data[0]

        flux_variance = aperture_photometry(
            data=self.images["uncertainty"] ** 2,
            apertures=self.aperture,
        )["aperture_sum"].data[0]

        if sky_subtract:
            self.sky_aperture = CircularAnnulus(
                positions, r_in=radius_sky_inner, r_out=radius_sky_outer
            )
            measured_flux_from_sky = (
                ApertureStats(self.images["data"], self.sky_aperture).median[0]
                * self.aperture.area
            )
            flux -= measured_flux_from_sky
            model_flux_from_sky = (
                ApertureStats(self.images["model"], self.sky_aperture).median[0]
                * self.aperture.area
            )
            flux_model -= model_flux_from_sky
            print(
                f"{measured_flux_from_sky:>10.1f} photons = 🌫 sky fluence in aperture"
            )
            print(
                f"{flux:>10.1f} photons = (⭐️ measured) - (🌫 sky) fluence in aperture"
            )

        print(f"{uncertainty:>10.1f} photons = 🃏 uncertainty on fluence in aperture")
        aperture_kw = dict(color="white", linewidth=2)
        sky_kw = dict(color="silver", linewidth=2, linestyle="--")

        text_kw = dict(
            va="top",
            ha="center",
            color="white",
        )

        if visualize:
            try:
                for a in self._plotted_apertures:
                    a[0].set_visible(False)
            except AttributeError:
                self._plotted_apertures = []
            # label the values
            plt.sca(self.imshowed["data"].axes)
            self._plotted_apertures.append(self.aperture.plot(**aperture_kw))
            if sky_subtract:
                self._plotted_apertures.append(self.sky_aperture.plot(**sky_kw))

            # plt.text(
            #    positions[0][0],
            #    positions[0][1] + 1.2 * radius,
            plt.title(
                f"measured flux in aperture:\n{flux:.1f}$\pm${uncertainty:.1f} {self.unit}"
            )
            #    **text_kw,
            # )

            plt.sca(self.imshowed["model"].axes)
            self._plotted_apertures.append(self.aperture.plot(**aperture_kw))
            if sky_subtract:
                self._plotted_apertures.append(self.sky_aperture.plot(**sky_kw))
            plt.title(
                f"model flux in aperture:\n{flux_model:.1f} {self.unit}",
            )

            plt.sca(self.imshowed["uncertainty"].axes)
            self._plotted_apertures.append(self.aperture.plot(**aperture_kw))
            if sky_subtract:
                self._plotted_apertures.append(self.sky_aperture.plot(**sky_kw))

            plt.title(
                f"uncertainty in aperture:\n{np.sqrt(flux_variance):.1f} {self.unit}"
            )

        return flux, flux_model, uncertainty

    def plot_photometry_with_radius(self):
        '''
        Plot how the photometry depends on aperture radius
        (ignoring sky subraction).
        '''
        flux, model, uncertainty = [], [], []
        radii = np.arange(0.2, 15, 0.2)
        for r in radii:
            fl, mo, un = self.do_photometry(
                radius=r, visualize=False, sky_subtract=False
            )
            flux.append(fl)
            model.append(mo)
            uncertainty.append(un)
        flux, model, uncertainty = (
            np.array(flux),
            np.array(model),
            np.array(uncertainty),
        )
        fi, ax = plt.subplots(
            2, 1, dpi=300, figsize=(6, 4), sharex=True, constrained_layout=True
        )

        fwhm_arcsec = self.psf.fwhm
        fwhm_pixels = fwhm_arcsec / self.pixel_scale

        def plot_fwhm():
            alpha = 0.5
            plt.axvline(fwhm_pixels.value, color="gray", alpha=alpha)
            plt.text(
                fwhm_pixels.value,
                0,
                f"  FWHM ({fwhm_arcsec} = {fwhm_pixels})\n",
                color="gray",
                alpha=alpha,
            )

        plt.sca(ax[0])
        plt.errorbar(radii, flux, uncertainty, label="measured")
        plt.plot(radii, model, label="model")
        plt.ylabel("Photons in Aperture")
        plot_fwhm()

        plt.legend(frameon=False)
        plt.ylim(0, None)
        plt.sca(ax[1])
        plt.plot(radii, flux / uncertainty, label="measured")
        plt.plot(radii, model / uncertainty, label="model")
        plt.xlabel("Aperture Radius (pixels)")
        plt.ylabel("S/N (flux/uncertainty)")
        plt.ylim(0, None)
        plot_fwhm()

        plt.legend(frameon=False)

    def __repr__(self):
        '''
        Represent this image as a string.
        '''
        return f"Image {self.shape} = {'+'.join([str(x) for x in self.components])}"

    def mu(self, t=1 * u.s):
        '''
        Calculate the expectation value for this image, for an exposure time. 
        '''
        return (
            np.sum(
                np.array(
                    [
                        getattr(x, f"model_mean_{self.type}")(t).to_value(self.unit)
                        for x in self.components
                    ],
                    dtype="object",
                ),
                axis=0,
            )
            * self.unit
        )

    def sigma(self, t=1 * u.s):
        '''
        Calculate the uncertainty for this image, for an exposure time. 
        '''
        return np.sqrt(
            np.sum(
                np.array(
                    [
                        getattr(x, f"model_sigma_{self.type}")(t).to_value(self.unit)
                        ** 2
                        for x in self.components
                    ],
                    dtype="object",
                ),
                axis=0,
            )
            * self.unit**2
        )

    def imshow_realization(
        self,
        t=1 * u.s,
        ax=None,
        filename="animation.mp4",
        photometry=False,
        photometry_positions=[(15, 15)],
        photometry_radius=3,
        **kwargs,
    ):
        '''
        Display a (noisy) realization of this image. 
        '''
        if ax is None:
            fi, ax = plt.subplots(1, 1, figsize=(4, 2.5), dpi=300)

        time = np.max(t)
        i = self.realization(t=time)
        e = self.sigma(t=time)
        imshowed = plt.imshow(i, **kwargs)
        plt.xticks([])
        plt.yticks([])
        plt.colorbar(label=self.unit.to_string("latex_inline"))

        if photometry:
            self.do_photometry(
                data=i.to_value(self.unit),
                uncertainty=e.to_value(self.unit),
                positions=photometry_positions,
                radius=photometry_radius,
            )

        if np.size(t) > 1:
            writer, displayer = get_animation_writer_and_displayer(filename)
            with writer.saving(fi, filename, fi.get_dpi()):
                for time in t:
                    i = self.realization(t=time)
                    imshowed.set_data(i)
                    writer.grab_frame()
            try:
                display(displayer(filename, embed=True))
            except TypeError:
                display(displayer(filename))

    def make_exposure(
        self,
        t=1 * u.s,
        visualize=True,
        ax=None,
        photometry=False,
        photometry_positions=[(15, 15)],
        photometry_radius=3,
        cmap="gray",
        **kwargs,
    ):
        '''
        Build up the model, uncertainty, and realized data for an exposure (and display).
        '''
        time = np.max(t)
        model = self.mu(t=time)
        uncertainty = self.sigma(t=time)
        data = self.realization(t=time)

        images = dict(
            data=data.to_value(self.unit),
            model=model.to_value(self.unit),
            uncertainty=uncertainty.to_value(self.unit),
        )
        if ax is None:
            fi, ax = plt.subplots(
                1,
                3 + 1,
                figsize=(10, 4),
                dpi=300,
                sharex=True,
                sharey=True,
                constrained_layout=True,
            )

        imshowed = {}
        for a, i, label in zip(
            ax,
            [model, uncertainty, data],
            [
                f"model",
                f"uncertainty",
                f"data",
            ],
        ):
            plt.sca(a)
            imshowed[label] = plt.imshow(i, cmap=cmap, **kwargs)
            plt.xticks([])
            plt.yticks([])
            plt.ylabel(label.capitalize())
            plt.colorbar(
                orientation="horizontal", label=self.unit.to_string("latex_inline")
            )

            if label == "data":
                if photometry:
                    self.do_photometry(
                        data=images["data"],
                        uncertainty=images["uncertainty"],
                        positions=photometry_positions,
                        radius=photometry_radius,
                    )

        plt.sca(ax[-1])
        texted = plt.text(
            0.1,
            0.0,
            f"exposure_time={time:.1f}\n" + self.describe_everything_as_string(),
            transform=plt.gca().transAxes,
            va="bottom",
            ha="left",
            fontsize=9,
            clip_on=False,
        )
        plt.axis("off")

        self.images = images
        self.imshowed = imshowed

    def imshow_with_models(
        self,
        t=1 * u.s,
        ax=None,
        filename="animation.mp4",
        photometry=False,
        photometry_positions=[(15, 15)],
        photometry_radius=3,
        **kwargs,
    ):
        time = np.max(t)
        model = self.mu(t=time)
        sigma = self.sigma(t=time)
        data = self.realization(t=time)

        if ax is None:
            fi, ax = plt.subplots(
                1,
                3 + 1,
                figsize=(13, 2.5),
                dpi=300,
                sharex=True,
                sharey=True,
                constrained_layout=True,
            )

        imshowed = {}
        for a, i, label in zip(
            ax,
            [data, model, sigma],
            [
                f"data",
                f"model",
                f"uncertainty",
            ],
        ):
            plt.sca(a)
            imshowed[label] = plt.imshow(i, **kwargs)
            plt.xticks([])
            plt.yticks([])
            plt.ylabel(label.capitalize())
            plt.colorbar(label=self.unit.to_string("latex_inline"))

            if label == "data":
                if photometry:
                    self.do_photometry(
                        data=data.to_value(u.adu),
                        uncertainty=sigma.to_value(u.adu),
                        positions=photometry_positions,
                        radius=photometry_radius,
                    )

        plt.sca(ax[-1])
        texted = plt.text(
            0.1,
            0.9,
            f"exposure_time={time:.1f}\n" + self.describe_everything_as_string(),
            transform=plt.gca().transAxes,
            va="top",
            ha="left",
            fontsize=9,
            clip_on=False,
        )
        plt.axis("off")

        return imshowed

        # FIXME -- allow for animations? 
        if np.size(t) > 1:
            writer, displayer = get_animation_writer_and_displayer(filename)
            with writer.saving(fi, filename, fi.get_dpi()):
                for time in t:
                    model = self.model_mean_uncalibrated(t=time)
                    sigma = self.model_sigma_uncalibrated(t=time)
                    data = self.realization(t=time)
                    for a, i, label in zip(
                        ax,
                        [data, model, sigma],
                        [
                            "Data (ADU)",
                            "Model ($\mu$, ADU)",
                            "Noise ($\sigma$, ADU)",
                        ],
                    ):
                        imshowed[label].set_data(i)
                    texted.set_text(
                        f"exposure_time={time:.1f}\n" + self.describe_as_string()
                    )
                    writer.grab_frame()
            try:
                display(displayer(filename, embed=True))
            except TypeError:
                display(displayer(filename))

    def imshow_components_with_models(self, **kwargs):
        '''
        
        '''
        rows = 4
        cols = len(self.components) + 2
        fi, ax = plt.subplots(
            rows,
            cols,
            figsize=(3 * cols, 2.5 * rows),
            dpi=300,
            sharex=True,
            sharey=True,
            constrained_layout=True,
            gridspec_kw=dict(width_ratios=[1, 0.25] + [1] * len(self.components)),
        )
        for col, component in enumerate(self.components):
            component.imshow_with_models(ax=ax[:, col + 2], **kwargs)
            for a in ax[:, col + 2]:
                a.set_title(component.__class__.__name__)

        self.imshow_with_models(ax=ax[:, 0], **kwargs)
        for a in ax[:, 1]:
            plt.sca(a)
            plt.text(
                0.5,
                0.5,
                "=",
                ha="left",
                va="center",
                fontsize=40,
                transform=a.transAxes,
            )
            plt.axis("off")
        for a in ax[:, 0]:
            a.set_title("$E_{light}$")

    def realization(self, *args, **kwargs):
        '''
        Create a noisy realization of this image. 
        '''
        mu = self.mu(*args, **kwargs).to_value(self.unit)
        sigma = self.sigma(*args, **kwargs).to_value(self.unit)
        return np.random.normal(mu, sigma) * self.unit


class CalibratedImage(Image):
    unit = u.photon
    type = "calibrated"

    """def mu(self, *args, **kwargs):
        return self.model_mean_calibrated(*args, **kwargs)

    def sigma(self, *args, **kwargs):
        return self.model_sigma_calibrated(*args, **kwargs)"""


class UncalibratedImage(Image):
    unit = u.adu
    type = "uncalibrated"

    """def mu(self, *args, **kwargs):
        return self.model_mean_uncalibrated(*args, **kwargs)

    def sigma(self, *args, **kwargs):
        return self.model_sigma_uncalibrated(*args, **kwargs)"""


class ImageComponent(Image):
    """
    An image Component, with a mean model and an uncertainty.
    """

    keys = []

    def describe_everything_as_string(self):
        return self.describe_as_string()

    def __init__(self, *args, **kwargs):
        Image.__init__(self, *args, **kwargs)

    def model_mean_calibrated(self, *args, **kwargs):
        return np.ones(self.image.shape) * u.photon

    def model_sigma_calibrated(self, *args, **kwargs):
        return np.zeros(self.image.shape) * u.photon

    def model_mean_uncalibrated(self, *args, **kwargs):
        return self.model_mean_calibrated(*args, **kwargs) / self.image.gain

    def model_sigma_uncalibrated(self, *args, **kwargs):
        return self.model_sigma_calibrated(*args, **kwargs) / self.image.gain

    def model_variance_photons(self, *args, **kwargs):
        return self.model_sigma_calibrated(*args, **kwargs) ** 2

    def model_variance_adu(self, *args, **kwargs):
        return self.model_sigma_uncalibrated(*args, **kwargs) ** 2

    def __repr__(self):
        keywords = [f"{k}={getattr(self, k)}" for k in self.keys]
        return f"{self.__class__.__name__}({', '.join(keywords)})"


class Bias(ImageComponent):
    keys = ["bias_level", "read_noise"]

    def __init__(self, bias_level=1000 * u.adu, read_noise=10 * u.electron, **kwargs):
        self.bias_level = bias_level
        self.read_noise = read_noise

        ImageComponent.__init__(self, **kwargs)

    def model_mean_calibrated(self, *args, **kwargs):
        return np.zeros(self.image.shape) * u.photon

    def model_sigma_calibrated(self, *args, **kwargs):
        if self.read_noise.unit == u.adu:
            sigma = self.read_noise * self.image.gain
        else:
            sigma = to_photons(self.read_noise)
        return np.ones(self.image.shape) * sigma

    def model_mean_uncalibrated(self, *args, **kwargs):
        return np.ones(self.image.shape) * self.bias_level

    def model_sigma_uncalibrated(self, *args, **kwargs):
        if self.read_noise.unit == u.adu:
            sigma = self.read_noise
        else:
            sigma = to_photons(self.read_noise) / self.image.gain
        return np.ones(self.image.shape) * sigma


class Dark(ImageComponent):
    keys = ["average_dark_rate"]

    def __init__(self, average_dark_rate=1 * u.photon / u.s, **kwargs):
        self.average_dark_rate = average_dark_rate
        ImageComponent.__init__(self, **kwargs)

    def initialize_dark(self):
        self._dark_rate = (
            np.random.exponential(
                self.average_dark_rate.to_value(u.photon / u.s), self.image.shape
            )
            * u.photon
            / u.s
        )

    @property
    def dark_rate(self):
        try:
            return self._dark_rate
        except AttributeError:
            self.initialize_dark()
            return self._dark_rate

    def model_mean_calibrated(self, t=1 * u.s):
        return np.zeros(self.image.shape) * self.image.unit

    def model_mean_uncalibrated(self, t=1 * u.s):
        return self.dark_rate * t / self.image.gain

    def model_sigma_calibrated(self, t=1 * u.s):
        photons = self.dark_rate * t
        return np.sqrt(photons * u.photon)


class Sky(ImageComponent):
    keys = ["sky_brightness"]

    def __init__(self, sky_brightness=3 * u.photon / u.s, **kwargs):
        self.sky_brightness = sky_brightness
        ImageComponent.__init__(self, **kwargs)

    def model_mean_calibrated(self, t=1 * u.s):
        return self.sky_brightness * np.ones(self.image.shape) * t

    def model_sigma_calibrated(self, t=1 * u.s):
        photons = self.sky_brightness * np.ones(self.image.shape) * t
        return np.sqrt(photons * u.photon)


class Star(ImageComponent):
    keys = ["positions", "brightness"]

    def __init__(
        self,
        positions=[(15, 15)],
        brightness=[1000] * u.photon / u.s,
        **kwargs,
    ):

        self.positions = np.atleast_2d(positions)
        self.brightness = np.atleast_1d(brightness)
        ImageComponent.__init__(self, **kwargs)

    @property
    def starlight_rate(self):
        try:
            return self._starlight_rate
        except AttributeError:
            self.initialize_stars()
            return self._starlight_rate

    def initialize_stars(self):

        x, y = np.meshgrid(
            np.arange(self.image.shape[1]), np.arange(self.image.shape[0])
        )
        N = len(self.positions)
        for i in range(N):
            x_center, y_center = self.positions[i]
            theta = (
                np.sqrt((x - x_center) ** 2 + (y - y_center) ** 2)
                * u.pixel
                * self.image.pixel_scale
            )
            pixel_area = (self.image.pixel_scale * u.pixel) ** 2
            this_star = self.brightness[i] * self.image.psf(theta) * pixel_area
            try:
                stars += this_star
            except NameError:
                stars = this_star
        self._starlight_rate = stars

    def model_mean_calibrated(self, t=1 * u.s):
        return self.starlight_rate * t

    def model_sigma_calibrated(self, t=1 * u.s):
        photons = self.starlight_rate * t
        return np.sqrt(photons * u.photon)



from ..imports import *
from ..gaia import get_gaia
from .apertures import *
from astropy.coordinates import SkyCoord 
from astropy import wcs
from astropy.io import fits
import astropy.units as u

from astropy.convolution import AiryDisk2DKernel, convolve, convolve_fft, Moffat2DKernel, CustomKernel
from astropy.visualization import imshow_norm, ManualInterval, MinMaxInterval, SqrtStretch, LogStretch


# FIXME, synthesize this with imaging.py
class SimulatedImage:
    def __init__(self, 
                 name='16 Cyg',
                 fov = 2*u.arcminute, 
                 pixel_size = 0.1*u.arcsecond, 
                 ):
        
        self.fov = fov 
        self.pixel_size = pixel_size 
        self.N = int((fov/pixel_size).decompose())

        self.setup_stars(name=name)
        self.setup_wcs()
        self.setup_pixels()

    def setup_stars(self, name):    
        self.name = name 
        self.stars = get_gaia(name, radius=self.fov)

    def setup_wcs(self):
        w = wcs.WCS(naxis=2)
        w.wcs.crpix = [self.N/2, self.N/2]
        center = self.stars.meta['center']
        w.wcs.cdelt = [self.pixel_size.to_value('deg'), self.pixel_size.to_value('deg')]
        w.wcs.crval = [center.ra.deg, center.dec.deg]
        w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
       
        self.wcs = w 
        self.header = w.to_header()

    def setup_pixels(self):
        '''
        Set up some pxiels, with stars as single pixels. 
        '''
        # calculate pixel locations for stars
        coordinates = SkyCoord(ra=self.stars['ra'], dec=self.stars['dec'])
        x, y = self.wcs.world_to_pixel(coordinates)
        mag = self.stars[f"G_gaia_mag"].to_value("mag")


        image = np.zeros((self.N,self.N))
        flux_zero_mag = 1e10
        star_flux = flux_zero_mag*10**(-0.4*mag)

        # make single-pixel stars
        row = y.astype(int)
        col = x.astype(int)
        ok = (row >= 0) * (row < self.N) * (col >= 0) * (col < self.N)
        image[row[ok], col[ok]] = star_flux[ok]

        self.image_single_pixels = image 

    def simulate(self, aperture=CircleAperture, read_noise=1, **psf_kw):
        psf_kw['dtheta'] = self.pixel_size
        self.aperture = aperture(**psf_kw)
        self.kernel = CustomKernel(self.aperture.psf)
        self.image_model = convolve_fft(self.image_single_pixels, self.kernel)
        noise = np.sqrt(self.image_model + read_noise**2)
        self.image = np.random.normal(self.image_model, noise)
        filename = 'test.fit'
        f = fits.ImageHDU(self.image, header=self.header)
        f.header['object'] = (self.name, 'where is the field centered?')
        #f.header['simulate'] = (title, 'what PSF was simulated?')
        f.header['scale'] = (self.pixel_size.to_value('arcsecond'), 'pixel scale in units of arcseconds/pixel')
        f.writeto(filename, overwrite=True)
        self.display()


    # FIXME!
    def display(self, title='Impossibly Good', kernel=None, vmin=1e-10, vmax=1e-3):
        
        extent = np.array([-self.N/2, self.N/2, -self.N/2, self.N/2])*self.pixel_size.to_value('arcsec')

        plt.figure()
        im, norm = imshow_norm(self.image, origin='lower',
                            interval=ManualInterval(vmin, vmax), 
                            stretch=LogStretch(), extent=extent)
        plt.xlabel('x [arcsec]')
        plt.ylabel('y [arcsec]')
        plt.colorbar()
        #plt.savefig(f'{title}.pdf')



"""
Some draft tools for CCD calibrations.
(You should probably ignore these;
check out `imaging.py` instead!)
"""
from .imports import *
from .psf import *


class Camera:
    def __init__(
        self,
        shape=(30, 50),
        bias_level=1000 * u.adu,
        average_dark_rate=1 * u.electron / u.s,
        gain=2 * u.electron / u.adu,
        read_noise=5 * u.adu,
        star_positions=[[12, 20]],
        star_brightnesses=[100] * u.electron / u.s,
        pixel_scale=0.7 * u.arcsec / u.pixel,
        sky_brightness=3 * u.electron / u.s,
        psf=seeing(fwhm=2.0 * u.arcsec),
    ):
        self.shape = shape
        self.bias_level = bias_level
        self.average_dark_rate = average_dark_rate
        self.initialize_dark()
        self.gain = gain
        self.read_noise = read_noise

        self.initialize_photons(
            star_positions=star_positions,
            star_brightnesses=star_brightnesses,
            sky_brightness=sky_brightness,
            pixel_scale=pixel_scale,
            psf=psf,
        )

    def initialize_photons(
        self,
        star_positions=[[12, 20]],
        star_brightnesses=[100] * u.electron / u.s,
        pixel_scale=0.7 * u.arcsec / u.pixel,
        sky_brightness=3 * u.electron / u.s,
        psf=seeing(fwhm=2.0 * u.arcsec),
    ):
        self.star_positions = np.atleast_2d(star_positions)
        self.star_brightnesses = star_brightnesses
        self.sky_brightness = sky_brightness
        self.pixel_scale = pixel_scale
        self.psf = psf
        x, y = np.meshgrid(np.arange(self.shape[1]), np.arange(self.shape[0]))
        N = len(star_positions)
        for i in range(N):
            x_center, y_center = star_positions[i]
            theta = (
                np.sqrt((x - x_center) ** 2 + (y - y_center) ** 2)
                * u.pixel
                * self.pixel_scale
            )
            pixel_area = (self.pixel_scale * u.pixel) ** 2
            this_star = star_brightnesses[i] * self.psf(theta) * pixel_area
            try:
                stars += this_star
            except NameError:
                stars = this_star
        self.star_photon_rate = stars
        self.sky_photon_rate = sky_brightness
        self.photon_rate = stars + sky_brightness

    def P_light_model(self, exposure_time=1 * u.s):
        return self.photon_rate * exposure_time

    def P_light_noise_model(self, exposure_time=1 * u.s):
        return self.make_Elight_noise_model(exposure_time) * self.gain

    def P_light(self, exposure_time=1 * u.s):
        mu = self.P_light_model(exposure_time)
        sigma = self.P_light_noise_model(exposure_time)
        return np.random.normal(mu, sigma)

    def initialize_dark(self):
        self.dark_rate = (
            np.random.exponential(
                self.average_dark_rate.to_value(u.electron / u.s), self.shape
            )
            * u.electron
            / u.s
        )

    def make_bias_model(self, *args, **kwargs):
        return np.ones(self.shape) * self.bias_level

    def make_dark_model(self, exposure_time=1 * u.s):
        return self.dark_rate * exposure_time / self.gain + self.make_bias_model()

    def make_light_model(self, exposure_time=1 * u.s):
        return (
            self.photon_rate / self.gain * exposure_time
            + self.dark_rate * exposure_time / self.gain
            + self.make_bias_model()
        )

    def make_bias_noise_model(self, *args, **kwargs):
        return np.ones(self.shape) * self.read_noise

    def make_dark_noise_model(self, exposure_time=1 * u.s):
        photons = self.dark_rate * exposure_time
        return np.sqrt(photons * u.electron) / self.gain

    def make_light_noise_model(self, exposure_time=1 * u.s):
        photons = self.photon_rate * exposure_time
        return np.sqrt(photons * u.electron) / self.gain

    def make_Elight_noise_model(self, exposure_time=1 * u.s):
        return np.sqrt(
            self.make_bias_noise_model() ** 2
            + self.make_dark_noise_model(exposure_time) ** 2
            + self.make_light_noise_model(exposure_time) ** 2
        )

    def make_bias_noise_instance(self, *args, **kwargs):
        return (
            np.random.normal(
                0, self.make_bias_noise_model(*args, **kwargs).to_value(u.adu)
            )
            * u.adu
        )

    def make_dark_noise_instance(self, *args, **kwargs):
        return (
            np.random.normal(
                0, self.make_dark_noise_model(*args, **kwargs).to_value(u.adu)
            )
            * u.adu
        )

    def make_light_noise_instance(self, *args, **kwargs):
        return (
            np.random.normal(
                0, self.make_light_noise_model(*args, **kwargs).to_value(u.adu)
            )
            * u.adu
        )

    def create_images(self, exposure_time=1 * u.s, include_noise=False):
        b = self.make_bias_model()
        d = self.make_dark_model(exposure_time=exposure_time)
        l = self.make_light_model(exposure_time=exposure_time)

        if include_noise:
            b_noise = self.make_bias_noise_instance(exposure_time=exposure_time)
            b += b_noise
            d_noise = self.make_dark_noise_instance(exposure_time=exposure_time)
            d += d_noise + b_noise
            l_noise = self.make_light_noise_instance(exposure_time=exposure_time)
            l += l_noise + d_noise + b_noise
        return b, d, l

    def imshow_exposures(
        self, exposure_time=1 * u.s, include_noise=False, vmin=None, vmax=None
    ):

        b, d, l = self.create_images(
            exposure_time=exposure_time, include_noise=include_noise
        )
        if vmin is None:
            vmin = np.min([b, d, l])
        if vmax is None:
            vmax = np.max([b, d, l])

        fi, ax = plt.subplots(
            1,
            4,
            figsize=(8, 3),
            dpi=600,
            sharex=True,
            sharey=True,
            constrained_layout=True,
        )
        for i, x, t in zip(
            range(3), [b, d, l], ["bias", "dark + bias", "light + dark + bias"]
        ):
            imshowed = ax[i].imshow(x, vmin=vmin, vmax=vmax, interpolation="nearest")
            ax[i].set_title(t)
            plt.colorbar(
                imshowed,
                orientation="horizontal",
                ax=ax[i],
                label="Pixel Brightness (ADU)",
            )
        for a in ax:
            a.set_xticks(np.arange(-0.5, self.shape[1], 1))
            a.set_yticks(np.arange(-0.5, self.shape[0], 1))
            a.set_xticklabels([])
            a.set_yticklabels([])
            a.grid(color="black", linewidth=1)

        a = ax[-1]
        plt.sca(a)
        a.axis("off")
        if include_noise:
            noise_string = f"""
with noise of...
{np.mean(self.make_bias_noise_model(exposure_time=exposure_time)):.1f} read noise
{np.mean(self.make_dark_noise_model(exposure_time=exposure_time)):.1f} dark noise on average
{np.max(self.make_light_noise_model(exposure_time=exposure_time)):.1f} photon noise at peak
..included"""

        else:
            noise_string = "with no noise included"
        s = f"""
exposure time = {exposure_time}

(rows, columns) = {self.shape}
gain = {self.gain}
average bias level = {self.bias_level}
average dark rate = {self.average_dark_rate}
brightest star = {np.max(self.star_brightnesses)}

{noise_string}"""
        plt.text(0, 0, s, transform=a.transAxes, ha="left", va="bottom", fontsize=6)

    def plot_pixels_with_exposure_time(
        self,
        exposure_times=np.linspace(0, 10) * u.s,
        include_noise=False,
        vmin=None,
        vmax=None,
    ):

        N = len(exposure_times)
        rows, cols = self.shape
        b_cube = np.zeros((N, rows, cols))
        d_cube = np.zeros((N, rows, cols))
        l_cube = np.zeros((N, rows, cols))
        for i, e in enumerate(exposure_times):
            b_cube[i, :, :], d_cube[i, :, :], l_cube[i, :, :] = self.create_images(
                exposure_time=e, include_noise=include_noise
            )

        for title in ["bias", "bias + dark", "bias + dark + light"]:
            fi, ax = plt.subplots(
                self.shape[0],
                self.shape[1],
                constrained_layout=True,
                dpi=600,
                sharex=True,
                sharey=True,
                figsize=(8, 6),
            )
            for i in range(rows):
                for j in range(cols):
                    ax[i, j].plot(
                        exposure_times,
                        b_cube[:, i, j],
                        label="bias",
                        linewidth=2,
                        alpha=0.5,
                        linestyle=":",
                    )
                    plt.ylim(0.95 * np.min(b_cube), 1.05 * np.max(b_cube))
                    if "dark" in title:
                        ax[i, j].plot(
                            exposure_times,
                            d_cube[:, i, j],
                            label="dark",
                            linewidth=2,
                            alpha=0.5,
                            linestyle="--",
                        )
                        plt.ylim(0.95 * np.min(b_cube), 1.05 * np.max(d_cube))
                    if "light" in title:
                        ax[i, j].plot(
                            exposure_times,
                            l_cube[:, i, j],
                            label="light",
                            linewidth=2,
                            alpha=0.5,
                            linestyle="-",
                        )
                        plt.ylim(0.95 * np.min(b_cube), 1.05 * np.max(l_cube))
            fi.suptitle(title)
            fi.supylabel("Pixel Brightness (ADU)")
            fi.supxlabel("Exposure Time (s)")
            plt.sca(ax[0, -1])
            plt.legend(frameon=False, bbox_to_anchor=(1, 1), loc="upper left")



from .imports import *
from astropy.uncertainty import uniform, normal, poisson
from astropy.visualization import quantity_support
from scipy import stats


def catch_photons_in_bucket(
    rate=1 * u.photon / u.s / u.m**2,
    diameter=2.0 * u.m/np.sqrt(np.pi),
    time=100 * u.s,
    visualize=True,
    figure=None,
):
    """
    Simulate a telescope catching photons,
    for a given photon flux (photons/s/m**2),
    telescope diameter (m), and exposure time (s).

    Parameters
    ----------
    rate: Quantity
        The rate at which photons are coming towards us,
        expressed in units of (photons/s/m**2).

    diameter: Quantity
        The effective diameter of a circular telescope,
        expressed in units of (m).

    time: Quantity
        The exposure duration, expressed in units of (s).

    Returns
    -------
    N: int
        The number of photons that land in the telescope
    """

    # what's the radius of the telescope
    radius = diameter / 2.0

    if visualize:
        with quantity_support():
            # make a rectangle that at least includes the telescope
            square = np.maximum(1.3 * diameter, 0.2 * u.m)

            # what's the expected total number of photons
            total_area = square * square
            area = np.pi * radius**2
            # create a rectangle of randomly located photons
            total_expectation = (total_area * time * rate).to_value(u.photon)
            expectation = (area * time * rate).to_value(u.photon)

            kw = dict(
                marker=".",
                markeredgecolor="none",
                alpha=0.5,
                linewidth=0,
                markersize=10,
            )
            if figure is None:
                figure = plt.figure(figsize=(10, 4), dpi=300)
            gs = plt.GridSpec(2, 2)
            plt.subplot(gs[:, 0])
            collected_color = "royalblue"

            Ntotal = poisson(total_expectation, n_samples=1).distribution[0]
            if Ntotal > 0:
                x = uniform(
                    lower=-square / 2, upper=square / 2, n_samples=Ntotal
                ).distribution
                y = uniform(
                    lower=-square / 2, upper=square / 2, n_samples=Ntotal
                ).distribution

                # determine which of these photons landed in the telescope
                incircle = (x**2 + y**2) < radius**2
                outofcircle = ~incircle
                N = np.sum(incircle)

                plt.plot(x[incircle], y[incircle], color=collected_color, **kw)
                plt.plot(x[outofcircle], y[outofcircle], color="black", **kw)
            else:
                N = 0
            # draw a circle
            theta = np.linspace(0, 2 * np.pi, 1000)
            plt.plot(
                radius * np.sin(theta),
                radius * np.cos(theta),
                linewidth=4,
                color=collected_color,
            )

            # label the radius of the circle
            plt.text(
                0,
                -radius * 1.2,
                f"{diameter.to_string(format='latex_inline', precision=4)}",
                ha="center",
                va="top",
                fontweight="bold",
                fontsize=10,
                color=collected_color,
            )
            plt.plot(
                radius * np.array([-1, 1]),
                -radius * 1.1 * np.ones(2),
                color=collected_color,
                linewidth=4,
            )

            # add a title
            plt.title(f"{N} photons")

            # set the aspect ratio of the plot to 1:1
            plt.axis("equal")
            plt.xlim(-square / 2, square / 2)
            plt.ylim(-square / 2, square / 2)

            # get rid of the square around the plot
            plt.axis("off")

            plt.subplot(gs[0, 1])
            kw = dict(format="latex_inline")
            plt.text(
                0,
                0,
                f"Photon Rate\n={rate.to_string(precision=4, **kw)}\n"
                + f"Telescope Diameter\n={diameter.to_string(precision=4, **kw)}\n"
                + f"Telescope Area\n={area.to_string(precision=4, **kw)}\n"
                + f"Exposure Time\n={time.to_string(precision=4, **kw)}\n"
                + f"Expected # of Photons\n={(expectation*u.photon).to_string(precision=4, **kw)}",
                transform=plt.gca().transAxes,
                ha="left",
                va="bottom",
            )
            plt.axis("off")
            plt.subplot(gs[1, 1])
            N_axis = np.arange(0, expectation + 5 * np.sqrt(expectation))
            PDF = stats.poisson(expectation).pmf(N_axis)
            plt.plot(N_axis, PDF, drawstyle="steps-mid", color="black")
            plt.axvline(N, color=collected_color, linewidth=4)
            plt.ylim(0, None)
            plt.xlabel("# of Photons")
            plt.ylabel("P(# of photons)")
    else:

        # if we don't need to make plot, just draw a Poisson number
        area = np.pi * radius**2
        expectation = (area * time * rate).to_value(u.photon)
        N = poisson(expectation, n_samples=1).distribution[0]

    return N



from ..imports import *

from scipy.special import jv
from astropy.visualization import *
from scipy.fft import fft2, fftfreq, fftshift
from scipy.signal import convolve2d

# from photutils.profiles import RadialProfile
import warnings


class Aperture:
    def __init__(self, **kw):
        self.make_aperture(**kw)
        self.make_psf(**kw)

    def __repr__(self):
        return self.__class__.__name__

    def label(self):
        return f"{self}-{self.wavelength}".replace(" ", "")

    def make_psf(self, wavelength=0.5 * u.micron, dtheta=None, supersample=5, **kw):

        self.wavelength = wavelength

        dx = np.median(np.diff(self.x))
        dy = np.median(np.diff(self.y))

        skip = 1
        if dtheta is None:
            N_y, N_x = np.array(np.shape(self.aperture)) * supersample
        else:
            N_x = int(wavelength / dx / dtheta.to_value("radian"))
            N_y = int(wavelength / dy / dtheta.to_value("radian"))
            if (N_x < 3) or (N_y < 3):
                print(
                    f"""
                The large pixel size of dtheta={dtheta} is 
                making it almost impossible to do the FFT. 
                Consider remaking your aperture with a 
                larger N=, so you'll still have resolution.
                """
                )
            if (N_x < len(self.x)) or (N_y < len(self.y)):
                skip = int(np.round(np.maximum(len(self.x) / N_x, len(self.y) / N_y)))
                N_x *= skip
                N_y *= skip
                print(
                    f"""
                Be careful! dtheta={dtheta} is so big the FFT is weird!
                Inflating by a factor of {skip}x.
                """
                )
            if N_x % 2 == 0:
                N_x += 1
            if N_y % 2 == 0:
                N_y += 1

        k_x = fftshift(fftfreq(N_x, dx))[::skip]
        k_y = fftshift(fftfreq(N_y, dy))[::skip]

        self.theta_x = (k_x * wavelength * u.radian).to("arcsec")
        self.theta_y = (k_y * wavelength * u.radian).to("arcsec")
        self.theta_x2d, self.theta_y2d = np.meshgrid(self.theta_x, self.theta_y)

        full_resolution_psf = np.abs(fftshift(fft2(self.aperture, (N_y, N_x)))) ** 2
        self.psf = convolve2d(full_resolution_psf, np.ones((skip, skip)), mode="same")[
            ::skip, ::skip
        ]

    def imshow_aperture_and_psf(self, scale="log", save=True, zoom=1):
        fi, ax = plt.subplots(1, 2, figsize=(8, 3), constrained_layout=True)
        self.imshow_aperture(
            ax=ax[0],
            save=False,
        )
        self.imshow_psf(ax=ax[1], save=False, zoom=zoom)
        plt.savefig(f"aperture+psf-imshow-{self.label()}.pdf")

    def display_psf(
        self, theta_range=10 * u.arcsec, stretches=["linear", "log"], save=True
    ):

        with quantity_support():
            span = theta_range.value
            unit = theta_range.unit

            fi, ax = plt.subplots(
                2,
                len(stretches),
                figsize=(8, 7),
                gridspec_kw=dict(height_ratios=[0.5, 1]),
                constrained_layout=True,
                sharex=True,
            )
            for i, scale in enumerate(stretches):

                self.plot_psf(
                    ax=ax[0, i], theta_range=theta_range, scale=scale, save=False
                )
                plt.xlabel("")

                self.imshow_psf(ax=ax[1, i], scale=scale, save=False, colorbar=False)
                if scale == "log":
                    label = r"log$_{10}$(PSF)"
                else:
                    label = "PSF"
                # if i > 0:
                #    ax[1, 0].get_shared_y_axes().join(ax[1, 0], ax[1, i])
                plt.colorbar(ax=ax[:, i], label=label, orientation="horizontal")
            if save:
                plt.savefig(f"aperture+psf-everything-{self.label()}.pdf")

    def imshow_aperture(self, ax=None, save=True, colorbar=True, **kw):
        with quantity_support():
            if ax is None:
                fi, ax = plt.subplots(1, 1, constrained_layout=True)
            plt.sca(ax)
            imshow_kw = dict(
                cmap="gray", extent=[min(self.x), max(self.x), min(self.y), max(self.y)]
            )
            imshow_kw.update(**kw)
            plt.imshow(self.aperture, **imshow_kw)
            plt.xlabel(
                r"$\sf x$ at Telescope Aperture"
                + f' ({self.x.unit.to_string("latex_inline")})'
            )
            plt.ylabel(
                r"$\sf y$ at Telescope Aperture"
                + f' ({self.y.unit.to_string("latex_inline")})'
            )
            if colorbar:
                plt.colorbar()

            if save:
                plt.savefig(f"aperture-imshow-{self.label()}.pdf")

    def imshow_psf(self, ax=None, scale="log", save=True, colorbar=True, zoom=1, **kw):
        with quantity_support():
            if ax is None:
                fi, ax = plt.subplots(1, 1, constrained_layout=True)
            plt.sca(ax)

            plt.cla()
            if scale == "log":
                image = np.log10(self.psf / np.nanmax(self.psf))
                vmin = -10
                vmax = 0
                image[np.isfinite(image) == False] = -6
            else:
                image = self.psf / np.nanmax(self.psf)
                vmin = 0
                vmax = 1
            imshow_kw = dict(
                cmap="gray",
                extent=[
                    min(self.theta_x.value),
                    max(self.theta_x.value),
                    min(self.theta_y.value),
                    max(self.theta_y.value),
                ],
                vmin=vmin,
                vmax=vmax,
            )
            imshow_kw.update(**kw)
            plt.imshow(image, **imshow_kw)
            plt.xlabel(
                r"$\sf \theta_x$ at Focal Plane"
                + f' ({self.theta_x.unit.to_string("latex_inline")})'
            )
            plt.ylabel(
                r"$\sf \theta_y$ at Focal Plane"
                + f' ({self.theta_y.unit.to_string("latex_inline")})'
            )

            plt.xlim(min(self.theta_x.value) / zoom, max(self.theta_x.value) / zoom)
            plt.ylim(min(self.theta_y.value) / zoom, max(self.theta_y.value) / zoom)

            if colorbar:
                if scale == "log":
                    label = r"log$_{10}$(PSF)"
                else:
                    label = "PSF"
                plt.colorbar(label=label)
            if save:
                plt.savefig(f"psf-imshow-{self.label()}.pdf")

    def plot_psf(self, ax=None, theta_range=10 * u.arcsec, scale="log", save=False):
        with quantity_support():
            if ax is None:
                fi, ax = plt.subplots(1, 1, constrained_layout=True)
            plt.sca(ax)

            N_x, N_y = len(self.theta_x), len(self.theta_y)
            i_x = int(np.interp(0 * u.arcsec, self.theta_x, np.arange(N_x)))
            i_y = int(np.interp(0 * u.arcsec, self.theta_y, np.arange(N_y)))

            if scale == "log":
                ymin = 1e-10
                ymax = 1
            else:
                ymin = 0
                ymax = 1

            # dtheta = np.median(np.diff(self.theta_x))
            # edge_radii = np.arange(np.min([N_x/2, N_y/2, theta_range/dtheta]))
            # radial_profile = RadialProfile(self.psf, [i_x, i_y], edge_radii)
            # plt.plot(radial_profile.radius, radial_profile.profile)

            psf = self.psf / np.nanmax(self.psf)
            # r = (self.theta_x2d**2 + self.theta_y2d**2).flatten()
            # i = np.argsort(r)
            # plt.plot(r[i], psf.flatten()[i], **kw)
            # plt.plot(-r[i], psf.flatten()[i], **kw)
            kw = dict(color="black")
            plt.plot(self.theta_x, psf[i_y, :], **kw)
            plt.yscale(scale)
            plt.ylim(ymin, None)
            plt.xlabel(
                r"$\sf \theta$ at Focal Plane"
                + f' ({self.theta_x.unit.to_string("latex_inline")})'
            )
            plt.ylabel("Relative Intensity")
            if save:
                plt.savefig(f"psf-slice-{self.label()}.pdf")

    def make_visualizations(self):
        self.imshow_aperture()
        self.imshow_psf()
        self.imshow_aperture_and_psf()
        self.display_psf()


example_apertures_directory = os.path.join(astr3510_directory, "psfs", "apertures")
default_aperture_image_filename = os.path.join(
    example_apertures_directory, "leavitt.jpg"
)


class BitmapAperture(Aperture):
    def __repr__(self):
        return f"{self.__class__.__name__}-{self.diameter}".replace(" ", "")

    def make_aperture(
        self, filename=default_aperture_image_filename, diameter=1 * u.m, **kw
    ):
        image = plt.imread(filename)
        if len(np.shape(image)) == 3:
            image = np.sum(image, -1)
        image = image / np.nanmax(image)
        rows, columns = np.shape(image)

        # set up a blank image
        self.diameter = diameter
        radius = diameter / 2
        aspect_ratio = columns / rows
        self.x = np.linspace(-radius, radius, columns) * aspect_ratio
        self.y = np.linspace(-radius, radius, rows)
        self.x2d, self.y2d = np.meshgrid(self.x, self.y)
        self.aperture = image


class JWST(BitmapAperture):
    def __repr__(self):
        return "JWST"

    def make_aperture(self, **kw):
        BitmapAperture.make_aperture(
            self,
            filename=os.path.join(example_apertures_directory, "jwst.jpg"),
            diameter=6.5 * u.m,
            **kw,
        )


class BahtinovAperture(BitmapAperture):
    def make_aperture(self, **kw):
        BitmapAperture.make_aperture(
            self,
            filename=os.path.join(example_apertures_directory, "bahtinov.jpg"),
            **kw,
        )


class ShapeAperture(Aperture):
    def make_aperture(self, radius=1 * u.m, N=201, buffer=1.1, **kw):

        # set up a blank image
        self.x = np.linspace(-radius, radius, N) * buffer
        self.y = np.linspace(-radius, radius, N) * buffer
        self.x2d, self.y2d = np.meshgrid(self.x, self.y)
        self.aperture = np.zeros((N, N))


class RectangleAperture(ShapeAperture):
    def __repr__(self):
        return f"Rectangle-{self.width}x{self.height}".replace(" ", "")

    def make_aperture(self, width=1 * u.m, height=1 * u.m, **kw):
        self.width = width
        self.height = height
        ShapeAperture.make_aperture(self, radius=np.maximum(width, height) / 2, **kw)
        r = np.sqrt(self.x2d**2 + self.y2d**2)
        self.aperture[
            (np.abs(self.x2d) < width / 2) * (np.abs(self.y2d) < height / 2)
        ] = 1


class CircleAperture(ShapeAperture):
    def __repr__(self):
        return f"Circle-{self.diameter}".replace(" ", "")

    def make_aperture(self, diameter=1 * u.m, **kw):
        self.diameter = diameter
        ShapeAperture.make_aperture(self, radius=diameter / 2, **kw)
        r = np.sqrt(self.x2d**2 + self.y2d**2)
        self.aperture[r < diameter / 2] = 1


class ObscuredCircleAperture(CircleAperture):
    def __repr__(self):
        return f"ObscuredCircle-diameter={self.diameter}-obscuration={self.obscuration_diameter}-spiders={self.spiders}".replace(
            " ", ""
        )

    def make_aperture(
        self,
        diameter=1 * u.m,
        obscuration_diameter=0.5 * u.m,
        spiders=4,
        spider_width=5 * u.cm,
        **kw,
    ):
        CircleAperture.make_aperture(self, diameter=diameter, **kw)
        self.obscuration_diameter = obscuration_diameter
        self.spiders = spiders
        self.spider_width = spider_width

        # substract central obscuration
        self.subtract_circle(diameter=obscuration_diameter)

        # subtract spiders
        self.subtract_spiders(spiders=spiders, spider_width=spider_width)

    def subtract_circle(self, diameter=0.5 * u.m, x_center=0 * u.m, y_center=0 * u.m):
        r = np.sqrt((self.x2d - x_center) ** 2 + (self.y2d - y_center) ** 2)
        self.aperture[r < diameter / 2] = 0

    def subtract_spiders(self, spiders=4, spider_width=5 * u.cm):
        r = np.sqrt((self.x2d) ** 2 + (self.y2d) ** 2)
        theta = np.arctan2(self.y2d, self.x2d)
        for spider_theta in np.linspace(0, 2 * np.pi, spiders + 1)[:-1] * u.radian:
            distance_from_spider = np.abs(np.sin(theta - spider_theta)) * r
            in_right_direction = np.cos(theta - spider_theta) > 0
            is_spider = (distance_from_spider < spider_width / 2) * in_right_direction
            self.aperture[is_spider] = 0


class Spitzer(ObscuredCircleAperture):
    def __repr__(self):
        return "Spitzer"

    def make_aperture(self, **kw):
        ObscuredCircleAperture.make_aperture(
            self,
            diameter=0.85 * u.m,
            obscuration_diameter=0.12 * u.m,
            spiders=3,
            spider_width=2 * u.cm,
            **kw,
        )


class Hubble(ObscuredCircleAperture):
    def __repr__(self):
        return "Hubble"

    def make_aperture(self, **kw):
        ObscuredCircleAperture.make_aperture(
            self,
            diameter=2.4 * u.m,
            obscuration_diameter=0.8 * u.m,
            spiders=4,
            spider_width=2 * u.cm,
            **kw,
        )

        # add weird mirror circles
        blob_distance_from_center = 0.9 * self.diameter / 2
        blob_diameter = 0.2 * u.m
        for blob_theta in np.array([30, 150, 270]) * np.pi / 180:
            self.subtract_circle(
                diameter=blob_diameter,
                x_center=blob_distance_from_center * np.cos(blob_theta),
                y_center=blob_distance_from_center * np.sin(blob_theta),
            )


class Leto(ObscuredCircleAperture):
    def __repr__(self):
        return "Leto"

    def make_aperture(self, **kw):
        ObscuredCircleAperture.make_aperture(
            self,
            diameter=0.61 * u.m,
            obscuration_diameter=0.2 * u.m,
            spiders=4,
            spider_width=2 * u.cm,
            **kw,
        )


class Artemis(ObscuredCircleAperture):
    def __repr__(self):
        return "Artemis"

    def make_aperture(self, **kw):
        ObscuredCircleAperture.make_aperture(
            self,
            diameter=0.508 * u.m,
            obscuration_diameter=0.198 * u.m,
            spiders=4,
            spider_width=2 * u.cm,
            **kw,
        )


class Airy(CircleAperture):
    """An analytic airy pattern!"""

    def airy(self, theta, wavelength=0.5 * u.micron, diameter=1 * u.m):
        k = 2 * np.pi / wavelength
        x = (k * diameter / 2 * np.sin(theta)).decompose()
        return ((2 * jv(1, x) / x) ** 2).decompose()

    def make_psf(self, wavelength=0.5 * u.micron, supersample=20, **kw):

        self.wavelength = wavelength

        N_y, N_x = np.array(np.shape(self.aperture)) * supersample
        dx = np.median(np.diff(self.x))
        dy = np.median(np.diff(self.y))
        k_x = fftshift(fftfreq(N_x, dx))
        k_y = fftshift(fftfreq(N_y, dy))

        self.theta_x = np.arcsin(k_x * wavelength).to("arcsec")
        self.theta_y = np.arcsin(k_y * wavelength).to("arcsec")
        self.theta_x2d, self.theta_y2d = np.meshgrid(self.theta_x, self.theta_y)

        theta = np.sqrt(self.theta_x2d**2 + self.theta_y2d**2)
        a = self.airy(theta, wavelength=self.wavelength, diameter=self.diameter)
        self.psf = a


apertures_to_test = [
    CircleAperture,
    ObscuredCircleAperture,
    RectangleAperture,
    JWST,
    BahtinovAperture,
    Hubble,
    Spitzer,
    Leto,
    Artemis,
]


"""class PSF:
    

"""


"""class seeing(PSF):
    def __init__(self, fwhm=1 * u.arcsecond):
        self.fwhm = fwhm

    def __call__(self, theta):
        sigma = self.fwhm / 2.354
        return 1 / (2 * np.pi * sigma**2) * np.exp(-0.5 * theta**2 / sigma**2)

    def __repr__(self):
        return f'{self.fwhm}" seeing'


class stellar_disk(PSF):
    def __init__(self, angular_radius=(1 * u.Rsun) / (10 * u.pc) * u.arcsecond):
        self.angular_radius = angular_radius

    def stellar_disk(theta):
        return (theta <= self.angular_radius) * 1.0
"""



import numpy as np


def exoplanet_transit(
    t,
    rp=0.1,
    per=3.0,
    t0=0.0,
    a=10.0,
    inc=90.0,
    ecc=0.0,
    w=0.0,
    u=[0.2, 0.2],
    *args,
    **kwargs,
):
    """
    One dimensional transit model with quadratic
    limb-darkening, using the `exoplanet-core` package.
    Parameters
    ----------
    t : array
        The times at which to evaluate the model.
    rp : float, array
        The radius of the planet in stellar radii
    per : float
        The period of the planet, in days.
    t0 : float
        Mid-transit time of the transit, in days.
    a : float, array
        The semi-major axis of the orbit, in stellar radii.
    inc : float, array
        The inclination of the orbit, in degrees.
    ecc: float, array
        The eccentricity of the orbit, unitless.
    w : float, array
        The argument of periastron, in degrees.
    u : array
        The quadratic limb-darkening coefficients.
    Returns
    -------
    monochromatic_flux : array
        The flux evaluated at each time.
    cached_inputs : dict
        A kludge to store any intermediate variables
        needed to speed up computation (= none for
        `exoplanet` transits).
    """

    try:
        from exoplanet_core import kepler, quad_limbdark_light_curve
    except ImportError:
        warnings.warn(
            f"""
        You're trying to produce a transit model using `exoplanet_core`,
        but it looks like you don't have `exoplanet_core` installed
        in the environment from which you're running `chromatic`.
        Please either install it with `pip install exoplanet-core`
        (see https://github.com/exoplanet-dev/exoplanet-core for details)
        or use the `.inject_transit(..., method='trapezoid')`
        option instead.
        """
        )

    # these two handy functions were adapted from the Exoplanet code:
    def warp_times(times, t_0, _pad=True):
        if _pad:
            return np.pad(t, (0, len(times)), "constant", constant_values=1) - t_0
        return times - t_0

    def get_true_anomaly(times, t_0, t_ref, n_per, e, _pad=True):
        M = (warp_times(times, t_0, _pad=_pad) - t_ref) * n_per
        if e == 0:
            return np.sin(M), np.cos(M)
        sin_f, cos_f = kepler(M, e + np.zeros(len(M)))
        return sin_f, cos_f

    n = 2 * np.pi / per
    opsw = 1 + np.sin(w)
    E0 = 2 * np.arctan2(
        np.sqrt(1 - ecc) * np.cos(w),
        np.sqrt(1 + ecc) * opsw,
    )
    M0 = E0 - ecc * np.sin(E0)
    t_periastron = t0 - M0 / n
    tref = t_periastron - t0

    # calculate the true anomaly as a function of time:
    sinf, cosf = get_true_anomaly(t, t0, tref, n, ecc, _pad=False)
    cosi = np.cos(inc * np.pi / 180)  # convert inclination to radians
    sini = np.sin(inc * np.pi / 180)  # convert inclination to radians
    w_rad = w * np.pi / 180  # convert omega to radians
    cos_w_plus_f = (np.cos(w_rad) * cosf) - (np.sin(w_rad) * sinf)  # cos(f+w)
    sin_w_plus_f = (np.sin(w_rad) * cosf) + (np.cos(w_rad) * sinf)  # sin(f+w)

    # x,y,z equations from Winn (2010):
    r = (a * (1 - (ecc**2))) / (1 + ecc * cosf)
    x = -r * cos_w_plus_f
    y = -r * sin_w_plus_f * cosi
    z = r * sin_w_plus_f * sini
    r_sky = np.sqrt(x**2 + y**2)

    # use exoplanet_core functions to extract light curve:
    # u_arr = np.array(u)
    flux = quad_limbdark_light_curve(u[0], u[1], r_sky, rp)
    # we only want the lightcurve where z > 0 (planet is between us and star)
    flux[z < 0] = 0
    return 1 + flux
